@@ -13,8 +13,12 @@ describe Rack::Lint do
     [200, { "content-type" => "test/plain", "content-length" => "3" }, ["foo"]]
   end
 
-  def env(*args)
-    Rack::MockRequest.env_for("/", *args)
+  def env(options = {})
+    unless options.key?(:input)
+      options[:input] = String.new
+    end
+
+    Rack::MockRequest.env_for("/", options)
   end
 
   it "pass valid request" do
@@ -22,137 +26,133 @@ describe Rack::Lint do
   end
 
   it "notice fatal errors" do
-    lambda { Rack::Lint.new(nil).call }.must_raise(Rack::Lint::LintError).
+    lambda { Rack::Lint.new(valid_app).call }.must_raise(Rack::Lint::LintError).
       message.must_match(/No env given/)
   end
 
   it "notice environment errors" do
-    lambda { Rack::Lint.new(nil).call 5 }.must_raise(Rack::Lint::LintError).
+    lambda { Rack::Lint.new(valid_app).call 5 }.must_raise(Rack::Lint::LintError).
       message.must_match(/not a Hash/)
 
-    lambda { Rack::Lint.new(nil).call({}.freeze) }.must_raise(Rack::Lint::LintError).
+    lambda { Rack::Lint.new(valid_app).call({}.freeze) }.must_raise(Rack::Lint::LintError).
       message.must_match(/env should not be frozen, but is/)
 
+    lambda { Rack::Lint.new(valid_app).call({a: 1}) }.must_raise(Rack::Lint::LintError).
+      message.must_equal("env contains non-string keys: [:a]")
+
     lambda {
       e = env
       e.delete("REQUEST_METHOD")
-      Rack::Lint.new(nil).call(e)
+      Rack::Lint.new(valid_app).call(e)
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/missing required key REQUEST_METHOD/)
 
     lambda {
       e = env
       e.delete("SERVER_NAME")
-      Rack::Lint.new(nil).call(e)
+      Rack::Lint.new(valid_app).call(e)
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/missing required key SERVER_NAME/)
 
     lambda {
       e = env
       e.delete("SERVER_PROTOCOL")
-      Rack::Lint.new(nil).call(e)
+      Rack::Lint.new(valid_app).call(e)
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/missing required key SERVER_PROTOCOL/)
 
     lambda {
       e = env
       e["SERVER_PROTOCOL"] = 'Foo'
-      Rack::Lint.new(nil).call(e)
+      Rack::Lint.new(valid_app).call(e)
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/env\[SERVER_PROTOCOL\] does not match HTTP/)
 
     lambda {
-      e = env
-      e["HTTP_VERSION"] = 'HTTP/1.0'
-      Rack::Lint.new(nil).call(e)
-    }.must_raise(Rack::Lint::LintError).
-      message.must_match(/env\[HTTP_VERSION\] does not equal env\[SERVER_PROTOCOL\]/)
-
-    lambda {
-      Rack::Lint.new(nil).call(env("HTTP_CONTENT_TYPE" => "text/plain"))
+      Rack::Lint.new(valid_app).call(env("HTTP_CONTENT_TYPE" => "text/plain"))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/contains HTTP_CONTENT_TYPE/)
 
     lambda {
-      Rack::Lint.new(nil).call(env("HTTP_CONTENT_LENGTH" => "42"))
+      Rack::Lint.new(valid_app).call(env("HTTP_CONTENT_LENGTH" => "42"))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/contains HTTP_CONTENT_LENGTH/)
 
     lambda {
-      Rack::Lint.new(nil).call(env("FOO" => Object.new))
+      Rack::Lint.new(valid_app).call(env("FOO" => Object.new))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/non-string value/)
 
     lambda {
-      Rack::Lint.new(nil).call(env("rack.url_scheme" => "gopher"))
+      Rack::Lint.new(valid_app).call(env("rack.url_scheme" => "gopher"))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/url_scheme unknown/)
 
     lambda {
-      Rack::Lint.new(nil).call(env("rack.session" => []))
+      Rack::Lint.new(valid_app).call(env("rack.session" => []))
     }.must_raise(Rack::Lint::LintError).
       message.must_equal "session [] must respond to store and []="
 
     Rack::Lint.new(valid_app).call(env("rack.session" => {}))[0].must_equal 200
 
     lambda {
-      Rack::Lint.new(nil).call(env("rack.session" => {}.freeze))
+      Rack::Lint.new(valid_app).call(env("rack.session" => {}.freeze))
     }.must_raise(Rack::Lint::LintError).
       message.must_equal "session {} must respond to to_hash and return unfrozen Hash instance"
 
     obj = {}
     obj.singleton_class.send(:undef_method, :to_hash)
     lambda {
-      Rack::Lint.new(nil).call(env("rack.session" => obj))
+      Rack::Lint.new(valid_app).call(env("rack.session" => obj))
     }.must_raise(Rack::Lint::LintError).
       message.must_equal "session {} must respond to to_hash and return unfrozen Hash instance"
 
     obj.singleton_class.send(:undef_method, :clear)
     lambda {
-      Rack::Lint.new(nil).call(env("rack.session" => obj))
+      Rack::Lint.new(valid_app).call(env("rack.session" => obj))
     }.must_raise(Rack::Lint::LintError).
       message.must_equal "session {} must respond to clear"
 
     obj.singleton_class.send(:undef_method, :delete)
     lambda {
-      Rack::Lint.new(nil).call(env("rack.session" => obj))
+      Rack::Lint.new(valid_app).call(env("rack.session" => obj))
     }.must_raise(Rack::Lint::LintError).
       message.must_equal "session {} must respond to delete"
 
     obj.singleton_class.send(:undef_method, :fetch)
     lambda {
-      Rack::Lint.new(nil).call(env("rack.session" => obj))
+      Rack::Lint.new(valid_app).call(env("rack.session" => obj))
     }.must_raise(Rack::Lint::LintError).
       message.must_equal "session {} must respond to fetch and []"
 
     obj = Object.new
     def obj.inspect; '[]' end
     lambda {
-      Rack::Lint.new(nil).call(env("rack.logger" => obj))
+      Rack::Lint.new(valid_app).call(env("rack.logger" => obj))
     }.must_raise(Rack::Lint::LintError).
       message.must_equal "logger [] must respond to info"
 
     def obj.info(*) end
     lambda {
-      Rack::Lint.new(nil).call(env("rack.logger" => obj))
+      Rack::Lint.new(valid_app).call(env("rack.logger" => obj))
     }.must_raise(Rack::Lint::LintError).
       message.must_equal "logger [] must respond to debug"
 
     def obj.debug(*) end
     lambda {
-      Rack::Lint.new(nil).call(env("rack.logger" => obj))
+      Rack::Lint.new(valid_app).call(env("rack.logger" => obj))
     }.must_raise(Rack::Lint::LintError).
       message.must_equal "logger [] must respond to warn"
 
     def obj.warn(*) end
     lambda {
-      Rack::Lint.new(nil).call(env("rack.logger" => obj))
+      Rack::Lint.new(valid_app).call(env("rack.logger" => obj))
     }.must_raise(Rack::Lint::LintError).
       message.must_equal "logger [] must respond to error"
 
     def obj.error(*) end
     lambda {
-      Rack::Lint.new(nil).call(env("rack.logger" => obj))
+      Rack::Lint.new(valid_app).call(env("rack.logger" => obj))
     }.must_raise(Rack::Lint::LintError).
       message.must_equal "logger [] must respond to fatal"
 
@@ -160,14 +160,14 @@ describe Rack::Lint do
     Rack::Lint.new(valid_app).call(env("rack.logger" => obj))[0].must_equal 200
 
     lambda {
-      Rack::Lint.new(nil).call(env("rack.multipart.buffer_size" => 0))
+      Rack::Lint.new(valid_app).call(env("rack.multipart.buffer_size" => 0))
     }.must_raise(Rack::Lint::LintError).
       message.must_equal "rack.multipart.buffer_size must be an Integer > 0 if specified"
 
     Rack::Lint.new(valid_app).call(env("rack.multipart.buffer_size" => 1))[0].must_equal 200
 
     lambda {
-      Rack::Lint.new(nil).call(env("rack.multipart.tempfile_factory" => Tempfile))
+      Rack::Lint.new(valid_app).call(env("rack.multipart.tempfile_factory" => Tempfile))
     }.must_raise(Rack::Lint::LintError).
       message.must_equal "rack.multipart.tempfile_factory must respond to #call"
 
@@ -187,52 +187,47 @@ describe Rack::Lint do
       message.must_equal "response array has 0 elements instead of 3"
 
     lambda {
-      Rack::Lint.new(nil).call(env("SERVER_PORT" => "howdy"))
+      Rack::Lint.new(valid_app).call(env("SERVER_PORT" => "howdy"))
     }.must_raise(Rack::Lint::LintError).
       message.must_equal 'env[SERVER_PORT] is not an Integer'
 
     lambda {
-      Rack::Lint.new(nil).call(env("SERVER_NAME" => "\u1234"))
+      Rack::Lint.new(valid_app).call(env("SERVER_NAME" => "A Host"))
     }.must_raise(Rack::Lint::LintError).
-      message.must_equal "\u1234 must be a valid authority"
+      message.must_equal "env[SERVER_NAME] must be a valid host"
 
     lambda {
-      Rack::Lint.new(nil).call(env("HTTP_HOST" => "\u1234"))
+      Rack::Lint.new(valid_app).call(env("HTTP_HOST" => "A Host"))
     }.must_raise(Rack::Lint::LintError).
-      message.must_equal "\u1234 must be a valid authority"
+      message.must_equal "env[HTTP_HOST] must be a valid authority"
 
     lambda {
-      Rack::Lint.new(nil).call(env("REQUEST_METHOD" => "FUCKUP?"))
+      Rack::Lint.new(valid_app).call(env("REQUEST_METHOD" => "FUCKUP?"))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/REQUEST_METHOD/)
 
     lambda {
-      Rack::Lint.new(nil).call(env("REQUEST_METHOD" => "OOPS?\b!"))
+      Rack::Lint.new(valid_app).call(env("REQUEST_METHOD" => "OOPS?\b!"))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/OOPS\?\\/)
 
     lambda {
-      Rack::Lint.new(nil).call(env("SCRIPT_NAME" => "howdy"))
+      Rack::Lint.new(valid_app).call(env("SCRIPT_NAME" => "howdy"))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/must start with/)
 
     lambda {
-      Rack::Lint.new(nil).call(env("PATH_INFO" => "../foo"))
-    }.must_raise(Rack::Lint::LintError).
-      message.must_match(/must start with/)
-
-    lambda {
-      Rack::Lint.new(nil).call(env("CONTENT_LENGTH" => "xcii"))
+      Rack::Lint.new(valid_app).call(env("CONTENT_LENGTH" => "xcii"))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/Invalid CONTENT_LENGTH/)
 
     lambda {
-      Rack::Lint.new(nil).call(env("QUERY_STRING" => nil))
+      Rack::Lint.new(valid_app).call(env("QUERY_STRING" => nil))
     }.must_raise(Rack::Lint::LintError).
       message.must_include('env variable QUERY_STRING has non-string value nil')
 
     lambda {
-      Rack::Lint.new(nil).call(env("QUERY_STRING" => "\u1234"))
+      Rack::Lint.new(valid_app).call(env("QUERY_STRING" => "\u1234"))
     }.must_raise(Rack::Lint::LintError).
       message.must_include('env variable QUERY_STRING has value containing non-ASCII characters and has non-ASCII-8BIT encoding')
 
@@ -244,29 +239,48 @@ describe Rack::Lint do
       e = env
       e.delete("PATH_INFO")
       e.delete("SCRIPT_NAME")
-      Rack::Lint.new(nil).call(e)
+      Rack::Lint.new(valid_app).call(e)
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/One of .* must be set/)
 
     lambda {
-      Rack::Lint.new(nil).call(env("SCRIPT_NAME" => "/"))
+      Rack::Lint.new(valid_app).call(env("PATH_INFO" => "", "SCRIPT_NAME" => ""))
+    }.must_raise(Rack::Lint::LintError).
+      message.must_match(/One of .* must be set/)
+
+    lambda {
+      e = env("PATH_INFO" => "")
+      e.delete("SCRIPT_NAME")
+      Rack::Lint.new(valid_app).call(e)
+    }.must_raise(Rack::Lint::LintError).
+      message.must_match(/One of .* must be set/)
+
+    lambda {
+      e = env("SCRIPT_NAME" => "")
+      e.delete("PATH_INFO")
+      Rack::Lint.new(valid_app).call(e)
+    }.must_raise(Rack::Lint::LintError).
+      message.must_match(/One of .* must be set/)
+
+    lambda {
+      Rack::Lint.new(valid_app).call(env("SCRIPT_NAME" => "/"))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/cannot be .* make it ''/)
 
     lambda {
-      Rack::Lint.new(nil).call(env("rack.response_finished" => "not a callable"))
+      Rack::Lint.new(valid_app).call(env("rack.response_finished" => "not a callable"))
     }.must_raise(Rack::Lint::LintError).
     message.must_match(/rack.response_finished must be an array of callable objects/)
 
     lambda {
-      Rack::Lint.new(nil).call(env("rack.response_finished" => [-> (env) {}, "not a callable"]))
+      Rack::Lint.new(valid_app).call(env("rack.response_finished" => [-> (env) {}, "not a callable"]))
     }.must_raise(Rack::Lint::LintError).
     message.must_match(/rack.response_finished values must respond to call/)
   end
 
   it "notice input errors" do
     lambda {
-      Rack::Lint.new(nil).call(env("rack.input" => ""))
+      Rack::Lint.new(valid_app).call(env("rack.input" => ""))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/does not respond to #gets/)
 
@@ -275,7 +289,7 @@ describe Rack::Lint do
       def input.binmode?
         false
       end
-      Rack::Lint.new(nil).call(env("rack.input" => input))
+      Rack::Lint.new(valid_app).call(env("rack.input" => input))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/is not opened in binary mode/)
 
@@ -288,7 +302,7 @@ describe Rack::Lint do
         end
         result
       end
-      Rack::Lint.new(nil).call(env("rack.input" => input))
+      Rack::Lint.new(valid_app).call(env("rack.input" => input))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/does not have ASCII-8BIT as its external encoding/)
   end
@@ -297,12 +311,19 @@ describe Rack::Lint do
     lambda {
       io = StringIO.new
       io.binmode
-      Rack::Lint.new(nil).call(env("rack.errors" => "", "rack.input" => io))
+      Rack::Lint.new(valid_app).call(env("rack.errors" => "", "rack.input" => io))
     }.must_raise(Rack::Lint::LintError).
       message.must_match(/does not respond to #puts/)
   end
 
   it "notice response errors" do
+    lambda {
+      Rack::Lint.new(lambda { |env|
+                       [200, {}, []].freeze
+                     }).call(env({}))
+    }.must_raise(Rack::Lint::LintError).
+      message.must_include('response is frozen')
+
     lambda {
       Rack::Lint.new(lambda { |env|
                        ""
@@ -318,6 +339,73 @@ describe Rack::Lint do
       message.must_include('response array has 4 elements instead of 3')
   end
 
+  it "accepts empty PATH_INFO" do
+    Rack::Lint.new(valid_app).call(env("PATH_INFO" => "", "SCRIPT_NAME" => "/foo")).first.must_equal 200
+  end
+
+  it "notices request-target asterisk form errors" do
+    # A non-empty PATH_INFO starting with something other than / has
+    # implications for Rack::Request#path and methods downstream from
+    # it. Note that RFC3875 does not actually anticipate dealing with
+    # `OPTIONS *`; that should be considered a bug in the spec.
+    Rack::Lint.new(valid_app).call(env("REQUEST_METHOD" => "OPTIONS", "PATH_INFO" => '*')).
+      first.must_equal 200
+
+    lambda do
+      Rack::Lint.new(valid_app).call(env("PATH_INFO" => "*"))
+    end.must_raise(Rack::Lint::LintError).
+      message.must_match(/Only OPTIONS requests may have PATH_INFO set to '\*'/)
+  end
+
+  it "notices request-target authority form errors" do
+    Rack::Lint.new(valid_app).call(env("REQUEST_METHOD" => "CONNECT", "PATH_INFO" => "example.com:80")).
+      first.must_equal 200
+
+    lambda do
+      Rack::Lint.new(valid_app).call(env("PATH_INFO" => "example.com:80"))
+    end.must_raise(Rack::Lint::LintError).
+      message.must_match(/Only CONNECT requests may have PATH_INFO set to an authority/)
+
+    lambda do
+      Rack::Lint.new(valid_app).call(env("PATH_INFO" => "/:80")).first.must_equal 200
+    end
+  end
+
+  it "notices request-target absolute-form errors" do
+    Rack::Lint.new(valid_app).call(env("REQUEST_METHOD" => "GET", "PATH_INFO" => "http://foo/bar")).
+      first.must_equal 200
+
+    lambda do
+      Rack::Lint.new(valid_app).call(env("REQUEST_METHOD" => "CONNECT", "PATH_INFO" => "http://foo/bar"))
+    end.must_raise(Rack::Lint::LintError).
+      message.must_match(/CONNECT and OPTIONS requests must not have PATH_INFO set to a URI/)
+
+    lambda do
+      Rack::Lint.new(valid_app).call(env("REQUEST_METHOD" => "OPTIONS", "PATH_INFO" => "http://foo/bar"))
+    end.must_raise(Rack::Lint::LintError).
+      message.must_match(/CONNECT and OPTIONS requests must not have PATH_INFO set to a URI/)
+  end
+
+  it "notices request-target origin-form errors" do
+    Rack::Lint.new(valid_app).call(env("REQUEST_METHOD" => "GET", "PATH_INFO" => "/foo/bar")).
+      first.must_equal 200
+
+    lambda do
+      Rack::Lint.new(valid_app).call(env("REQUEST_METHOD" => "GET", "PATH_INFO" => "../etc/passwd"))
+    end.must_raise(Rack::Lint::LintError).
+      message.must_match(/PATH_INFO must start with a '\/'/)
+
+    lambda do
+      Rack::Lint.new(valid_app).call(env("REQUEST_METHOD" => "GET", "PATH_INFO" => "/foo/bar#qux"))
+    end.must_raise(Rack::Lint::LintError).
+      message.must_match(/PATH_INFO.*must not include a fragment/)
+
+    lambda do
+      Rack::Lint.new(valid_app).call(env("REQUEST_METHOD" => "GET", "PATH_INFO" => "/foo/bar?baz#qux"))
+    end.must_raise(Rack::Lint::LintError).
+      message.must_match(/PATH_INFO.*must not include a fragment/)
+  end
+
   it "notice status errors" do
     lambda {
       Rack::Lint.new(lambda { |env|
@@ -436,6 +524,61 @@ describe Rack::Lint do
 
   end
 
+  it "notice rack.early_hints errors" do
+    def self.env(arg={})
+      super(arg.merge("rack.early_hints" => proc{}))
+    end
+    def self.app(value, env={})
+      app = Rack::Lint.new(lambda { |env|
+                             env['rack.early_hints'].call(value)
+                             [200, {}, []]
+                           })
+      lambda { app.call(self.env.merge(env)) }
+    end
+
+    app({}, { 'rack.early_hints' => Object.new }).must_raise(Rack::Lint::LintError).
+      message.must_equal "rack.early_hints must respond to call"
+
+    app(Object.new).must_raise(Rack::Lint::LintError).
+      message.must_equal "headers object should be a hash, but isn't (got Object as headers)"
+
+    app({}.freeze).must_raise(Rack::Lint::LintError).
+      message.must_equal "headers object should not be frozen, but is"
+
+    app(true => false).must_raise(Rack::Lint::LintError).
+      message.must_equal "header key must be a string, was TrueClass"
+
+    app("status" => "404").must_raise(Rack::Lint::LintError).
+      message.must_match(/must not contain status/)
+
+    invalid_headers = 0.upto(31).map(&:chr) + %W<( ) , / : ; < = > ? @ [ \\ ] { } \x7F>
+    invalid_headers.each do |invalid_header|
+      app(invalid_header => "text/plain").
+        must_raise(Rack::Lint::LintError, "on invalid header: #{invalid_header}").
+        message.must_equal("invalid header name: #{invalid_header}")
+    end
+
+    ('A'..'Z').each do |invalid_header|
+      app(invalid_header => "text/plain").
+        must_raise(Rack::Lint::LintError, "on invalid header: #{invalid_header}").
+        message.must_equal("uppercase character in header name: #{invalid_header}")
+    end
+
+    valid_headers = 0.upto(127).map(&:chr) - invalid_headers - ('A'..'Z').to_a
+    valid_headers.each do |valid_header|
+      app(valid_header => "text/plain").call.first.must_equal 200
+    end
+
+    app("foo" => Object.new).must_raise(Rack::Lint::LintError).
+      message.must_equal "a header value must be a String or Array of Strings, but the value of 'foo' is a Object"
+
+    app("foo-bar" => "text\000plain").must_raise(Rack::Lint::LintError).
+      message.must_match(/invalid header/)
+
+    app([%w(content-type text/plain), %w(content-length 0)]).must_raise(Rack::Lint::LintError).
+      message.must_equal "headers object should be a hash, but isn't (got Array as headers)"
+  end
+
   it "notice content-type errors" do
     # lambda {
     #   Rack::Lint.new(lambda { |env|
@@ -483,6 +626,17 @@ describe Rack::Lint do
     body.to_path.must_equal __FILE__
   end
 
+  it "handles body.to_path returning nil" do
+    body = Object.new
+    def body.each; end
+    def body.to_path; nil end
+    app = lambda { |env| [200, {}, body] }
+
+    body = Rack::Lint.new(app).call(env({}))[2]
+    body.must_respond_to(:to_path)
+    body.to_path.must_be_nil
+  end
+
   it "notice body errors" do
     lambda {
       body = Rack::Lint.new(lambda { |env|
@@ -508,7 +662,16 @@ describe Rack::Lint do
       body.each { |part| }
       body.each { |part| }
     }.must_raise(Rack::Lint::LintError).
-      message.must_equal 'Response body must only be invoked once (each)'
+      message.must_equal 'Response body must only be called once (each)'
+
+    lambda {
+      body = Rack::Lint.new(lambda { |env|
+                               [200, { "content-type" => "text/plain", "content-length" => "0" }, []]
+                             }).call(env({}))[2]
+      body.close
+      body.each { |part| }
+    }.must_raise(Rack::Lint::LintError).
+      message.must_equal 'Response body is already closed'
 
     # Lint before and after the Rack middleware being tested.
     def stacked_lint(app)
@@ -605,7 +768,18 @@ describe Rack::Lint do
                              }).call(env({}))[2]
       body.each { |part| }
     }.must_raise(Rack::Lint::LintError).
-      message.must_equal 'The file identified by body.to_path does not exist'
+      message.must_equal 'body.to_path must be nil or a path to an existing file'
+
+    lambda {
+      body = Rack::Lint.new(lambda { |env|
+                               to_path = Object.new
+                               def to_path.each; end
+                               def to_path.to_path; false end
+                               [200, { "content-type" => "text/plain", "content-length" => "0" }, to_path]
+                             }).call(env({}))[2]
+      body.each { |part| }
+    }.must_raise(Rack::Lint::LintError).
+      message.must_equal 'body.to_path must be nil or a path to an existing file'
 
     lambda {
       body = Rack::Lint.new(lambda { |env|
@@ -622,7 +796,16 @@ describe Rack::Lint do
       body.call(StringIO.new)
       body.call(nil)
     }.must_raise(Rack::Lint::LintError).
-      message.must_equal 'Response body must only be invoked once (call)'
+      message.must_equal 'Response body must only be called once (call)'
+
+    lambda {
+      body = Rack::Lint.new(lambda { |env|
+                               [200, { "content-type" => "text/plain", "content-length" => "0" }, proc{}]
+                             }).call(env({}))[2]
+      body.close
+      body.call(StringIO.new)
+    }.must_raise(Rack::Lint::LintError).
+      message.must_equal 'Response body is already closed'
 
     lambda {
       body = Rack::Lint.new(lambda { |env|
@@ -826,6 +1009,60 @@ describe Rack::Lint do
       message.must_match(/rack.hijack must respond to call/)
   end
 
+  it "handles valid rack.hijack env" do
+    begin
+      hijack_called = false
+      s = File.open(__FILE__, 'rb')
+      env = env({ 'rack.hijack' => proc { |io| hijack_called = true; s } })
+      res = Rack::Lint.new(lambda { |env|
+                       [201, { "content-type" => "text/plain", "content-length" => "0"}, []]
+                     }).call(env)
+      hijack_called.must_equal false
+      env['rack.hijack'].call.must_be_same_as s
+      hijack_called.must_equal true
+    ensure
+      s&.close
+    end
+  end
+
+  it "notices when rack.hijack env entry does not respond to #call" do
+    lambda {
+      Rack::Lint.new(lambda { |env|
+                       [201, { "content-type" => "text/plain", "content-length" => "0" }, []]
+                     }).call(env({ 'rack.hijack' => Object.new}))
+    }.must_raise(Rack::Lint::LintError).
+      message.must_equal 'rack.hijack must respond to call'
+  end
+
+  it "notices when rack.hijack env entry does not return an IO" do
+    env = env({ 'rack.hijack' => proc { Object.new } })
+    app = Rack::Lint.new(lambda { |env|
+                          [201, { "content-type" => "text/plain", "content-length" => "0" }, []]
+                         }).call(env)
+    env['rack.hijack'].must_raise(Rack::Lint::LintError).
+      message.must_equal 'rack.hijack must return an IO instance'
+  end
+
+  it "handles valid rack.hijack response header" do
+    hijack_called = false
+    res = Rack::Lint.new(lambda { |env|
+                     [201, { "content-type" => "text/plain", "content-length" => "0",
+                             'rack.hijack' =>  proc { |io| hijack_called = true; io.write('1') }}, []]
+                   }).call(env({ 'rack.hijack?' => true }))
+    hijack_called.must_equal false
+    s = StringIO.new
+    res[1]['rack.hijack'].call(s)
+    s.rewind
+    s.read.must_equal '1'
+    hijack_called.must_equal true
+  end
+
+  it "allows non-hijack responses when server supports hijacking" do
+    Rack::Lint.new(lambda { |env|
+                       [201, { "content-type" => "text/plain", "content-length" => "0"}, []]
+                     }).call(env({ 'rack.hijack?' => true }))
+  end
+
   it "notices when the response headers don't have a valid rack.hijack callback" do
     lambda {
       Rack::Lint.new(lambda { |env|
@@ -835,6 +1072,15 @@ describe Rack::Lint do
       message.must_equal 'rack.hijack header must respond to #call'
   end
 
+  it "notices when the response headers has a rack.hijack callback with hijacking being supported" do
+    lambda {
+      Rack::Lint.new(lambda { |env|
+                       [201, { "content-type" => "text/plain", "content-length" => "0", 'rack.hijack' =>  Object.new }, []]
+                     }).call(env({}))
+    }.must_raise(Rack::Lint::LintError).
+      message.must_equal 'rack.hijack header must not be present if server does not support hijacking'
+  end
+
   it "pass valid rack.response_finished" do
     callable_object = Class.new do
       def call(env, status, headers, error)
@@ -845,4 +1091,50 @@ describe Rack::Lint do
                      [200, {}, ["foo"]]
                    }).call(env({ "rack.response_finished" => [-> (env) {}, lambda { |env| }, callable_object], "content-length" => "3" })).first.must_equal 200
   end
+
+  it "notices when the response protocol is not an array of strings" do
+    app = Rack::Lint.new(lambda{|env|
+      [101, {'rack.protocol' => 'websocket'}, ["foo"]]
+    })
+
+    lambda do
+      response = app.call(env({'rack.protocol' => 'websocket'}))
+    end
+      .must_raise(Rack::Lint::LintError)
+      .message.must_equal("rack.protocol must be an Array of Strings")
+  end
+
+  it "notices when the response protocol is specified in the response but not in the request" do
+    app = Rack::Lint.new(lambda{|env|
+      [101, {'rack.protocol' => 'websocket'}, ["foo"]]
+    })
+
+    lambda do
+      app.call(env())
+    end
+      .must_raise(Rack::Lint::LintError)
+      .message.must_match(/rack.protocol header is "websocket", but rack.protocol was not set in request/)
+  end
+
+  it "notices when the response protocol is specified in the response but not in the request" do
+    app = Rack::Lint.new(lambda{|env|
+      [101, {'rack.protocol' => 'websocket'}, ["foo"]]
+    })
+
+    lambda do
+      app.call(env('rack.protocol' => ['smtp']))
+    end
+      .must_raise(Rack::Lint::LintError)
+      .message.must_match(/rack.protocol header is "websocket", but should be one of \["smtp"\] from the request!/)
+  end
+
+  it "pass valid rack.protocol" do
+    app = Rack::Lint.new(lambda{|env|
+      [101, {'rack.protocol' => 'websocket'}, ["foo"]]
+    })
+
+    response = app.call(env({'rack.protocol' => ['websocket']}))
+
+    response.first.must_equal 101
+  end
 end
