@@ -1,7 +1,7 @@
 # frozen_string_literal: true
 
 require_relative 'helper'
-require 'cgi'
+require 'cgi/escape'
 require 'forwardable'
 require 'securerandom'
 
@@ -578,6 +578,10 @@ class RackRequestTest < Minitest::Spec
     req = make_request \
       Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "example.org", "HTTP_FORWARDED" => "host=example.com:9292", "SERVER_PORT" => "9393")
     req.host_with_port.must_equal "example.com:9292"
+
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "example.org", "HTTP_X_FORWARDED_HOST" => "")
+    req.host_with_port.must_equal "example.org"
   end
 
   it "parse the query string" do
@@ -621,13 +625,8 @@ class RackRequestTest < Minitest::Spec
         super(){|h, k| h[k.to_s] if k.is_a?(Symbol)}
       end
     end
-    parser = Rack::QueryParser.new(c, 100)
-    c = Class.new(Rack::Request) do
-      define_method(:query_parser) do
-        parser
-      end
-    end
-    req = c.new(Rack::MockRequest.env_for("/?foo=bar&quux=bla"))
+    req = Rack::Request.new(Rack::MockRequest.env_for("/?foo=bar&quux=bla"))
+    req.query_parser = Rack::QueryParser.new(c, 100)
     req.GET[:foo].must_equal "bar"
     req.GET[:quux].must_equal "bla"
     req.params[:foo].must_equal "bar"
@@ -778,7 +777,7 @@ class RackRequestTest < Minitest::Spec
   it "safely accepts POST requests with empty body" do
     mr = Rack::MockRequest.env_for("/",
       "REQUEST_METHOD" => "POST",
-      "CONTENT_TYPE"   => "multipart/form-data, boundary=AaB03x",
+      "CONTENT_TYPE"   => "multipart/form-data; boundary=AaB03x",
       "CONTENT_LENGTH" => '0',
       :input => nil)
 
@@ -796,21 +795,133 @@ class RackRequestTest < Minitest::Spec
     req.POST.must_equal "foo" => "bar", "quux" => "bla"
   end
 
-  it "return values for the keys in the order given from values_at" do
-    req = make_request Rack::MockRequest.env_for("?foo=baz&wun=der&bar=ful")
+  it "return form_pairs for url-encoded POST data" do
+    req = make_request \
+      Rack::MockRequest.env_for("/",
+        'REQUEST_METHOD' => 'POST', :input => "foo=bar&quux=bla")
+    req.form_pairs.must_equal [["foo", "bar"], ["quux", "bla"]]
+  end
 
-    assert_output(nil, /deprecated/) do
-      req.values_at('foo').must_equal ['baz']
-      req.values_at('foo', 'wun').must_equal ['baz', 'der']
-      req.values_at('bar', 'foo', 'wun').must_equal ['ful', 'baz', 'der']
-    end
+  it "preserve duplicate keys in form_pairs" do
+    req = make_request \
+      Rack::MockRequest.env_for("/",
+        'REQUEST_METHOD' => 'POST', :input => "foo=1&foo=2&bar=3")
+    req.form_pairs.must_equal [["foo", "1"], ["foo", "2"], ["bar", "3"]]
+  end
 
-    next if self.class == TestProxyRequest
+  it "handle empty values in form_pairs" do
+    req = make_request \
+      Rack::MockRequest.env_for("/",
+        'REQUEST_METHOD' => 'POST', :input => "foo=&bar=baz&empty")
+    req.form_pairs.must_equal [["foo", ""], ["bar", "baz"], ["empty", nil]]
+  end
 
-    capture_warnings(req) do |warnings|
-      req.values_at('foo').must_equal ['baz']
-      warnings.pop.must_equal ["Request#values_at is deprecated and will be removed in a future version of Rack. Please use request.params.values_at instead", { uplevel: 1 }]
-    end
+  it "return empty array for form_pairs with no POST data" do
+    req = make_request \
+      Rack::MockRequest.env_for("/", 'REQUEST_METHOD' => 'POST', :input => "")
+    req.form_pairs.must_equal []
+  end
+
+  it "return empty array for form_pairs with non-form content type" do
+    req = make_request \
+      Rack::MockRequest.env_for("/",
+        'REQUEST_METHOD' => 'POST',
+        "CONTENT_TYPE" => 'text/plain',
+        :input => "foo=bar")
+    req.form_pairs.must_equal []
+  end
+
+  it "raise same error for form_pairs as POST with invalid encoding" do
+    req = make_request \
+      Rack::MockRequest.env_for("/",
+        'REQUEST_METHOD' => 'POST', :input => "a%=1")
+    lambda { req.form_pairs }.must_raise(Rack::Utils::InvalidParameterError).
+      message.must_equal "invalid %-encoding (a%)"
+  end
+
+  it "return form_pairs for multipart form data" do
+    input = <<EOF
+--AaB03x\r
+content-disposition: form-data; name="reply"\r
+\r
+yes\r
+--AaB03x\r
+content-disposition: form-data; name="name"\r
+\r
+John\r
+--AaB03x--\r
+EOF
+    req = make_request Rack::MockRequest.env_for("/",
+                      "CONTENT_TYPE" => "multipart/form-data; boundary=AaB03x",
+                      "CONTENT_LENGTH" => input.size,
+                      :input => input)
+
+    pairs = req.form_pairs
+    pairs.must_equal [["reply", "yes"], ["name", "John"]]
+  end
+
+  it "preserve duplicate keys in multipart form_pairs" do
+    input = <<EOF
+--AaB03x\r
+content-disposition: form-data; name="item"\r
+\r
+first\r
+--AaB03x\r
+content-disposition: form-data; name="item"\r
+\r
+second\r
+--AaB03x\r
+content-disposition: form-data; name="other"\r
+\r
+value\r
+--AaB03x--\r
+EOF
+    req = make_request Rack::MockRequest.env_for("/",
+                      "CONTENT_TYPE" => "multipart/form-data; boundary=AaB03x",
+                      "CONTENT_LENGTH" => input.size,
+                      :input => input)
+
+    pairs = req.form_pairs
+    pairs.must_equal [["item", "first"], ["item", "second"], ["other", "value"]]
+  end
+
+  it "include file uploads in multipart form_pairs" do
+    input = <<EOF
+--AaB03x\r
+content-disposition: form-data; name="reply"\r
+\r
+yes\r
+--AaB03x\r
+content-disposition: form-data; name="fileupload"; filename="test.txt"\r
+content-type: text/plain\r
+\r
+file content\r
+--AaB03x--\r
+EOF
+    req = make_request Rack::MockRequest.env_for("/",
+                      "CONTENT_TYPE" => "multipart/form-data; boundary=AaB03x",
+                      "CONTENT_LENGTH" => input.size,
+                      :input => input)
+
+    pairs = req.form_pairs
+    pairs.length.must_equal 2
+    pairs[0].must_equal ["reply", "yes"]
+    pairs[1][0].must_equal "fileupload"
+    pairs[1][1].must_be_kind_of Hash
+    pairs[1][1][:filename].must_equal "test.txt"
+    pairs[1][1][:type].must_equal "text/plain"
+  end
+
+  it "return empty array for empty multipart form_pairs" do
+    input = <<EOF
+--AaB03x--\r
+EOF
+    req = make_request Rack::MockRequest.env_for("/",
+                      "CONTENT_TYPE" => "multipart/form-data; boundary=AaB03x",
+                      "CONTENT_LENGTH" => input.size,
+                      :input => input)
+
+    req.form_pairs.must_equal []
   end
 
   it "extract referrer correctly" do
@@ -851,24 +962,6 @@ class RackRequestTest < Minitest::Spec
     req.media_type.must_be_nil
   end
 
-  deprecated "cache, but invalidates the cache" do
-    req = make_request \
-      Rack::MockRequest.env_for("/?foo=quux",
-        "CONTENT_TYPE" => "application/x-www-form-urlencoded",
-        :input => "foo=bar&quux=bla")
-    req.GET.must_equal "foo" => "quux"
-    req.GET.must_equal "foo" => "quux"
-    req.set_header("QUERY_STRING", "bla=foo")
-    req.GET.must_equal "bla" => "foo"
-    req.GET.must_equal "bla" => "foo"
-
-    req.POST.must_equal "foo" => "bar", "quux" => "bla"
-    req.POST.must_equal "foo" => "bar", "quux" => "bla"
-    req.set_header("rack.input", StringIO.new("foo=bla&quux=bar"))
-    req.POST.must_equal "foo" => "bla", "quux" => "bar"
-    req.POST.must_equal "foo" => "bla", "quux" => "bar"
-  end
-
   it "figure out if called via XHR" do
     req = make_request(Rack::MockRequest.env_for(""))
     req.wont_be :xhr?
@@ -1215,7 +1308,7 @@ content-transfer-encoding: base64\r
 --AaB03x--\r
 EOF
     req = make_request Rack::MockRequest.env_for("/",
-                      "CONTENT_TYPE" => "multipart/form-data, boundary=AaB03x",
+                      "CONTENT_TYPE" => "multipart/form-data; boundary=AaB03x",
                       "CONTENT_LENGTH" => input.size,
                       :input => input)
 
@@ -1254,7 +1347,7 @@ content-transfer-encoding: base64
 --AaB03x--
 EOF
     req = make_request Rack::MockRequest.env_for("/",
-                      "CONTENT_TYPE" => "multipart/form-data, boundary=AaB03x",
+                      "CONTENT_TYPE" => "multipart/form-data; boundary=AaB03x",
                       "CONTENT_LENGTH" => input.size,
                       :input => input)
 
@@ -1278,7 +1371,7 @@ content-transfer-encoding: base64\r
 --AaB03x--\r
 EOF
     req = make_request Rack::MockRequest.env_for("/",
-                      "CONTENT_TYPE" => "multipart/form-data, boundary=AaB03x",
+                      "CONTENT_TYPE" => "multipart/form-data; boundary=AaB03x",
                       "CONTENT_LENGTH" => input.size,
                       :input => input)
 
@@ -1309,7 +1402,7 @@ EOF
 EOF
     mr = Rack::MockRequest.env_for(
       "/",
-      "CONTENT_TYPE" => "multipart/form-data, boundary=AaB03x",
+      "CONTENT_TYPE" => "multipart/form-data; boundary=AaB03x",
       "CONTENT_LENGTH" => input.size,
       :input => input
     )
@@ -1392,7 +1485,7 @@ content-disposition: form-data; name="mean"; filename="mean"\r
 --AaB03x--\r
 EOF
     req = make_request Rack::MockRequest.env_for("/",
-                      "CONTENT_TYPE" => "multipart/form-data, boundary=AaB03x",
+                      "CONTENT_TYPE" => "multipart/form-data; boundary=AaB03x",
                       "CONTENT_LENGTH" => input.size,
                       :input => input)
 
@@ -1418,7 +1511,7 @@ content-transfer-encoding: base64\r
 --AaB03x--\r
 EOF
     env = Rack::MockRequest.env_for("/",
-                          "CONTENT_TYPE" => "multipart/form-data, boundary=AaB03x",
+                          "CONTENT_TYPE" => "multipart/form-data; boundary=AaB03x",
                           "CONTENT_LENGTH" => input.size,
                           :input => input)
     req = make_request(env)
@@ -1432,7 +1525,7 @@ EOF
 content-disposition: form-data; name="huge"; filename="huge"\r
 EOF
     req = make_request Rack::MockRequest.env_for("/",
-                      "CONTENT_TYPE" => "multipart/form-data, boundary=AaB03x",
+                      "CONTENT_TYPE" => "multipart/form-data; boundary=AaB03x",
                       "CONTENT_LENGTH" => input.size,
                       :input => input)
 
@@ -1445,7 +1538,7 @@ content-disposition: form-data; name="huge"; filename="huge"\r
 foo\r
 EOF
     req = make_request Rack::MockRequest.env_for("/",
-                      "CONTENT_TYPE" => "multipart/form-data, boundary=AaB03x",
+                      "CONTENT_TYPE" => "multipart/form-data; boundary=AaB03x",
                       "CONTENT_LENGTH" => input.size,
                       :input => input)
 
@@ -1458,7 +1551,7 @@ content-disposition: form-data; name="huge"; filename="huge"\r
 foo\r
 EOF
     req = make_request Rack::MockRequest.env_for("/",
-                      "CONTENT_TYPE" => "multipart/form-data, boundary=AaB03x",
+                      "CONTENT_TYPE" => "multipart/form-data; boundary=AaB03x",
                       "CONTENT_LENGTH" => input.size,
                       :input => input)
 
@@ -1471,7 +1564,7 @@ EOF
 content-disposition: form-data; name="huge"; filename="huge"\r
 EOF
     req = make_request Rack::MockRequest.env_for("/",
-                      "CONTENT_TYPE" => "multipart/form-data, boundary=AaB03x",
+                      "CONTENT_TYPE" => "multipart/form-data; boundary=AaB03x",
                       "CONTENT_LENGTH" => input.size,
                       :input => input)
 
@@ -1490,7 +1583,7 @@ foo\r
 --AaB03x--\r
 EOF
     req = make_request Rack::MockRequest.env_for("/",
-                      "CONTENT_TYPE" => "multipart/related, boundary=AaB03x",
+                      "CONTENT_TYPE" => "multipart/related; boundary=AaB03x",
                       "CONTENT_LENGTH" => input.size,
                       :input => input)
 
@@ -1508,7 +1601,7 @@ content-type: application/octet-stream\r
 EOF
 
         req = make_request Rack::MockRequest.env_for("/",
-                          "CONTENT_TYPE" => "multipart/form-data, boundary=AaB03x",
+                          "CONTENT_TYPE" => "multipart/form-data; boundary=AaB03x",
                           "CONTENT_LENGTH" => input.size,
                           :input => input)
 
@@ -1556,7 +1649,7 @@ content-transfer-encoding: base64\r
 EOF
     input.force_encoding(Encoding::ASCII_8BIT)
     res = Rack::MockRequest.new(Rack::Lint.new(app)).get "/",
-      "CONTENT_TYPE" => "multipart/form-data, boundary=AaB03x",
+      "CONTENT_TYPE" => "multipart/form-data; boundary=AaB03x",
       "CONTENT_LENGTH" => input.size.to_s, "rack.input" => StringIO.new(input)
 
     res.must_be :ok?
@@ -1927,7 +2020,7 @@ EOF
       def_delegators :@req, :env, :has_header?, :get_header, :fetch_header,
         :each_header, :set_header, :add_header, :delete_header
 
-      def_delegators :@req, :[], :[]=, :values_at
+      def_delegators :@req, :[], :[]=
 
       def initialize(req)
         @req = req
